	LIST P=PIC16F819, F=INHX8M
;	ERRORLEVEL  2
	__CONFIG _CP_OFF & _WDT_OFF & _XT_OSC & _LVP_OFF & _MCLR_OFF

;*****************************************************************************
;
; BATCAP: A program to determine the capacity of a battery.
;
; Program	BATCAP.ASM
; Author	Wim de Vree, pe1grl@amsat.org
; Revision history
;    20030715	V 0.3	Created initial program
;    20030717	V 0.4	Added "Discharging" bar during discharge	
;    20030720	V 0.5	Rewrite with floating point library
;			You now specify the resistor load, the
;			current is calculated (I = U/R)
;    20030722	V 0.7	Added Serial Routines and storage of voltage
;			every 10 minutes	
;			** FIRST LIVE VERSION **
;    20030727	V 0.8	Fixed floating point bug
;			Initial code for setup added
;    20030803	V 0.9	Sampling interval can be user defined
;			(1/2/4/5/6/10/15/20/30 mins are nice
;			 values, the others don't wrap round) 
;    20041024		Now keeps up to 128 interval samples
;			using BANK1 and BANK2
;    20041025		Rewritten to use Vref input with a 2.5V
;			reference diode
;    20041026		Added menu specification of input range
;			(2.5 / 5.0 /7.5 and 15 V in new setup)
;    20041027	V 1.1	Sampling interval is now selected from
;			a table to avoid illegal values
;			allowed are: 5,10,15 and 20 minutes
;    20041030	V 1.2	Some rewiring, Added background
;			illumination for 4 seconds if a button
;			is pressed
;    20041102	V 1.3	Added low-bat detection check on startup
;			(Uses pin AN1)
;    20041106	V 1.4	Load resistor can now be specified in .5 ohm
;			steps, range 1..120 ohms
;    20041108	V 1.5	Interval is now autoranging in intervals of 5
;			10 and 20 minutes, not user specified
;			Also transmit the capacity over the serial interface
;    20041109	V 1.6	When a discharge is completed save all the data into
;			EE, Pressing the Adjust/Send Button during the welcome
;			message will retransmit all this data

	include <p16f819.inc>
	include <../inc/macros2.inc>

;#define	LCD_ONE_LINE	1	; Comment this for a 16*2 LCD display

;*****************************************************************************
; Equates, I/O, vars
;*****************************************************************************
RESET_V		EQU	0x0000		; Address of RESET Vector
ISR_V		EQU	0x0004		; Address of Interrupt Vector
CLOCKFREQ	EQU	.40000

#define		pRelais		PORTB,1
#define		LCD_RS		PORTB,2
#define		LCD_ENABLE	PORTB,3
#define		LCD_DATA4	PORTB,4
#define		LCD_DATA5	PORTB,5
#define		LCD_DATA6	PORTB,6
#define		LCD_DATA7	PORTB,7

#define		LCD_BackLight	PORTA,2
#define		But1Sense	PORTA,5
#define		But2Sense	PORTA,4

#define		But1State	DBNC_cva,0
#define		But2State	DBNC_cva,1

#define		E_But1Pressed	EVENT,0
#define		E_But2Pressed	EVENT,1
#define		E_Tick		EVENT,2 
#define		E_Second	EVENT,3	; 1 second flag
#define		E_Minute	EVENT,5	; 1 minute flag
#define		E_SendSerial	EVENT,6	; Serial Send requested

; Variable declarations, DO NOT CHANGE THE ORDER OF THE DECLARATION !!!!!!
; This is used when variables are load/saved from EE memory

	CBLOCK 0x20			; Declare variable addresses starting at 0x50 
	    EVENT			; Contains various Event flags
	    TICKS1,TICKS2		; fraction timers 
	    SECS,MINS,HOURS		; Time keeping vars
	    B2A_RESULT: 5		; Byte to ASCII conversion
	    DBNC_VAR: 4			; Switch debouncing subroutine
	    SCRATCH: 8			; Scratch buffer for subroutines
	    FPVAR: .12			; Used in Floating point library
	    HiCount, LoCount
	    cva_old			; Used for generating events on Down/Up sequence
	    Resistor			; Load Resistor in ohms * 2
	    CO_Voltage			; Cutoff Voltage 0..255
	    Interval			; Interval for storing Discharge Voltages
	    V_Range			; Voltage range, 0 = 2.5V / 1 = 5.0V
	    ActVoltage			; Actual voltage 0..255
	    VScale: 3			; Voltage Multiplier to get mV range
	    Current: 3			; Current in mA
	    Capacity: 3			; Battery capacity in mAmin (FLOAT)
	    SampleCnt			; nr of samples taken for discharge curve
	    TEMP,TEMP2			; some TEMP's used in the main program
	    DischargeIcon		; Counter to display bar icons
	    ScreenNr			; Screen to display in main loop
	    BacklightCnt		; Secs to keep backlight on
        ENDC 

	CBLOCK 0xA0
	    SampleArrLow:.64		; 1st half Discharge Curve is kept here
	ENDC
	CBLOCK 0x120
	    SampleArrHigh:.64		; 2nd half Discharge Curve is kept here
	ENDC

_w		EQU	0x7E		; Save area for 'W' register
_status		EQU	0x7F		; Save area for 'STATUS' register

;*****************************************************************************
; Program start
;*****************************************************************************
	ORG	RESET_V			; RESET vector location
RESET		GOTO	START

;*****************************************************************************
; This is the Periperal Interrupt routine.
;*****************************************************************************
	ORG	ISR_V			; Interrupt vector location
INTERRUPT	GOTO	INTHANDLER

	include <../inc/bin2asc.inc>
	include <../inc/debounce.inc>

	CONSTANT	LCD_LINE0 = 0x000
	CONSTANT	LCD_LINE1 = 0x040

	include <../inc/lcd4p.inc>
	include <../inc/lcdcommon.inc>
	
#define	SERPORT		PORTB		; Port used for serial output
#define	TXBIT		0		; Bit used for serial output
#define	BDRATE		.4800		; requested Baud Rate
#define	MAXIM_USED	1		; MAX232 level converter used
	include <../inc/serout.inc>

	include <../inc/div168.inc>
	include	<../inc/fplib24.inc>
	
;*****************************************************************************
; Initialize processor registers
;*****************************************************************************
START					; POWER_ON Reset (Beginning of program)
		clrf	STATUS		; Do initialization, Select bank 0
		clrf	INTCON		; Clear int-flags, Disable interrupts
		clrf	PCLATH		; Keep in lower 2KByte

		movlw	0x20		; Clear memory in BANK0
		movwf	FSR
_memclear	clrf	INDF
		incf	FSR,f
		movlw	0x7F
		subwf	FSR,w
		bnz	_memclear

		clrf	ADCON0
		
		clrf	PORTA		; ALL PORT output should output Low.
		clrf	PORTB
		bcf	pRelais		; Active high (MOSFET, not a relais)

		BANK1
		movlw	b'00111011'	; RA2 output, all others inputs
		movwf	TRISA
		movlw	b'00000000'	; RB0-7 outputs
		movwf	TRISB
					; Setup A/D convertor
		movlw	b'00000101'	; 2 input channel + Vref, left justified, clk/2 disabled
		movwf	ADCON1		;

		bsf	OPTION_REG, NOT_RBPU
					; Disable PORTB pull-ups
		BANK0

		movlw	b'01000001'	; Fosc/8, Channel 0, A/D Converter O
		movwf	ADCON0

		movlw	.10
		movwf	TICKS1
		movlw	.5
		movwf	TICKS2

		clrf	T2CON		; Stop Timer2, Prescaler = 1:1, Postscaler = 1:1
		clrf	TMR2		; Clear Timer2 register
		clrf	INTCON		; Disable interrupts
		BANK1
		clrf	PIE1		; Disable peripheral interrupts
		BANK0	
		clrf	PIR1		; Clear peripheral interrupts Flags
					; Configure timer 2 to generate a 100 Hz
					; interrupt:
					; CLK/4 	 = 1.000.000 Hz
					; Prescaler 1:16 =    62.500 Hz
					; PR2 value 250  =       250 Hz
					; Postscaler 1:5 =        50 Hz
		movlw	b'00100010'	; Prescaler = 1:16, Postscaler 1:5
;		movlw	b'00100001'	; DEBUG TEST Prescaler = 1:4, Postscaler 1:5 (speed things up)

		movwf	T2CON		; Timer2 is off
		BANK1
		movlw	.249		; Timer 2 counts to 250 (0..249)
		movwf	PR2
		bsf	PIE1,TMR2IE	; Timer 2 interrupt on 
		BANK0
		bsf	INTCON,PEIE	; Pheriperal interrupts enabled
		bsf	INTCON,GIE	; Global interrupts enabled
		bsf	T2CON, TMR2ON	; Timer2 starts to increment
		clrf	EVENT
		
INIT:		call	LCDINIT		; Initialize LCDisplay
		movlw	b'00000100'	; Display on, Cursor off
		call	LCDDMODE 
		call	SERINIT		; Serial TX line idle
		call	LoadPars	; Retreive last used settings from EE
		bsf	FPFLAGS,RND	; enable FP rounding
		
		movlw	LOW msgVersion	; Display message on top line
		call	ShowMsg

_initloop	btfsc	E_But2Pressed	; On button 2 send historical data
		 call	SendEEData	; if this is available
		btfss	E_But1Pressed
		 goto	_initloop	; Wait for button 1 press
		bcf	E_But1Pressed

		bsf	LCD_BackLight	; Switch LCD Backlight on during param input

; Initally do a supply power check to prevent the battery from running
; out during the test

		call	CheckPower	; Check internal voltage
		sublw	.230		; equals 4.5 V
		skpc
		 goto	PowerOK		; Supply Battery is OK

		movlw	LOW msgLowBat
		call	ShowMsg
		btfss	E_But1Pressed
		 goto	$-1		; Wait for keypress
		bcf	E_But1Pressed

; The Voltage Range should be the first parameter to enter as it calculates
; the scaling used for the input of the Cut-off Voltage

PowerOK:	movlw	LOW msgSetVRange
		call	ShowMsg
		call	GetVRange	; Set the Voltage Range
		bcf	E_But1Pressed

		movlw	LOW msgSetResistor
		call	ShowMsg
		call	GetResistor	; Set the Resistor load
		bcf	E_But1Pressed

		movlw	LOW msgSetCutOff
		call	ShowMsg
		call	GetCOVoltage	; Set the CutOff voltage
		bcf	E_But1Pressed

		call	SavePars	; Save Current settings in EE
		
		movlw	LOW msgShowGo	; Display message on top line
		call	ShowMsg
		
WaitStart:	btfsc	E_But1Pressed
		 goto	StartDischarge
		call	LCDHOME
		call	Measure
		call	DispVoltage
		btfss	E_Tick		; 10 updates/sec max
		 goto	$-1
		bcf	E_Tick
		goto	WaitStart
		 
StartDischarge:	bcf	LCD_BackLight	; Switch LCD Backlight off
		bsf	pRelais		; Start Discharging
		call	LCDCLEAR
		call	LoadBarChars	; Load Discharing icons in char map

		clrf	EVENT		; Clear ALL events
		clrf	SECS		; init Timer to 00:00:00
		clrf	MINS
		clrf	HOURS

		clrf	SampleCnt	; No samples yet
		movlw	.5		; Start with a 5 minute interval
		movwf	Interval	; for sample storage

		FPLDC24	.100		; Dirty hack to clear Capacity, Preload with .01
		call	FPMOV24		; Doesn't work with 0.00 (Why ?)
		FPLDC24	.1
		call	FPDIV24	
		FPSTF24	Capacity

		call	Measure
		call 	StoreVoltage	; Save sample at start of discharge
				
DischargeLoop
;	ifdef	LCD_ONE_LINE
		btfss	E_But1Pressed
		 goto	_DL1
		call	CheckBackLight
		bcf	E_But1Pressed	; Key1 press, toggle display
		incf	ScreenNr,f
		call	LCDCLEAR
;	endif

_DL1		btfss	E_But2Pressed
		 goto	_DL2
		call	CheckBackLight

		bcf	E_But2Pressed	; Key 2 press, send collected data
		call	SendSamples

_DL2		btfss	E_Second	; Has a second expired ?
		 goto	DischargeLoop

		bcf	E_Second	; Clear the second flag
		;
		; Do the stuff which should happen every second:
		;
		; 1) Read the Battery voltage and display the current value
		; 2) Update the discharge bar
		; 3) Update the LCD Background illumination status
		;
		movf    BacklightCnt,W
		skpz
		 call	UpdateBackLight

		call	LCDHOME

		incf	DischargeIcon,w	; Update to the next icon
		andlw	0x7
		movwf	DischargeIcon
		xorlw	0x7		; Shrinking bar
		call	LCDPUTCHAR	; And show the icon
		
		movlw	2
		call	LCDSDDA
		
		call	Measure
		call	CalcCurrent
		
;	ifdef	LCD_ONE_LINE
		btfss	ScreenNr,0
		 call	DispScreen1
		btfsc	ScreenNr,0
		 call	DispScreen2
;	else
;		call	DispScreen
;	endif

		movf	ActVoltage,w	; If register >= constant, jump.
		subwf	CO_Voltage,w
		skpnc
		 goto	StopDischarge	; Battery is empty, stop discharge cycle

		btfss	E_Minute	; Has a minute expired ?
		 goto	DischargeLoop
		bcf	E_Minute

		; If we get here a complete minute has gone by, if the minute
		; counter is a multiple of the interval add a new voltage to
		; the discharge curve.
		; Up to 128 samples can be stored so the complete discharge
		; cycle should not take longer than 32 hours.

		call	AdjustCapacity	; Adjust with current * 1 min

		clrf	__dvdHI		; Sample the voltage at specified intervals
		movf	MINS,w
		movwf	__dvdLO
		movfw	Interval
		movwf	__dvsrLO
		call	DIV168
		movf	__dvdHI,w
		skpnz
		 call	StoreVoltage	; Add a value to the discharge curve
		GOTO	DischargeLoop	; Stay here forever

StopDischarge:	call	StoreVoltage	; Save stop value for discharge curve
		bcf	pRelais		; Remove the loading resistor
		movlw	LOW msgShowRdy	; Display message on top line
		call	ShowMsg
		call	DispCapacity
		call	SaveDischarge	; Save Discharge Data in EE

hangup		call	CheckBackLight	
		btfsc	E_But1Pressed
		 bcf	E_But1Pressed	; Key 1 press, not used here
		 
		btfss	E_But2Pressed
		 goto	_h1
		bcf	E_But2Pressed	; Key 2 press, send collected data
		call	SendSamples
_h1		btfss	E_Second	; Has a second expired ?
		 goto	hangup
		bcf	E_Second	; Clear the second flag
		movf    BacklightCnt,W
		skpz
		 call	UpdateBackLight	; Update the LCD Background illumination status
		goto	hangup

CheckBackLight	btfsc	E_But1Pressed
		 goto	_BL_on
		btfsc	E_But2Pressed
		 goto	_BL_on
		return			; No buttons are pressed
_BL_on		movlw	.6
		movwf	BacklightCnt	; Secs to keep backlight on
		call	UpdateBackLight
		return

UpdateBackLight	decfsz	BacklightCnt,F
		 goto	_blON
_blOFF		bcf	LCD_BackLight	; Switch LCD Backlight off
		return
_blON		bsf	LCD_BackLight	; Switch LCD Backlight on
		return

;*****************************************************************************
; StoreVoltage keeps the discharge curve in BANK1 & BANK2
; 128 values are kept here: 64 in BANK1 and 64 in BANK2
; It also takes care of the auto-interval adjustment: if the array
; is full subsample the data and double the interval.
;*****************************************************************************
StoreVoltage:	movlw	.128		; Max 128 samples (no more mem available)
		subwf	SampleCnt,w
		bnz	_doStore
		movlw	.20		; Array full: double sample interval
		subwf	Interval,w
		skpnz
		 return			; Max interval reached, do not store anymore

		; Subsample the data set, keep all elements on 0/2/4 and place
		; them on location 0/1/2 etc. Because the data is kept in two banks
		; it is done in two steps.

		movlw	.32
		movwf	SCRATCH		; Scratch keeps the counter

		movlw	SampleArrLow	; Copy first 32 samples
		movwf	SCRATCH+1	; Source Pointer
		movwf	SCRATCH+2	; Destination Pointer
_subsample1	movfw	SCRATCH+1
		movwf	FSR		; Points to source
		movfw	INDF
		movwf	TEMP		; Value now in TEMP
		movfw	SCRATCH+2
		movwf	FSR		; Point to destination
		movfw	TEMP
		movwf	INDF		; Value now in array
		incf	SCRATCH+1,F
		incf	SCRATCH+1,F	; srcptr = srcptr + 2
		incf	SCRATCH+2,F	; dstptr = dstptr + 1
		decfsz	SCRATCH,F	; loop 32 samples
		 goto	_subsample1

		movlw	.32
		movwf	SCRATCH		; Copy second 32 samples
		movlw	low SampleArrHigh 
		movwf	SCRATCH+1	; Source Pointer
_subsample2	movfw	SCRATCH+1
		movwf	FSR		; Points to source
		bsf	STATUS,IRP	; source is in BANK3
		movfw	INDF
		bcf	STATUS,IRP
		movwf	TEMP		; Value now in TEMP
		movfw	SCRATCH+2
		movwf	FSR		; Point to destination
		movfw	TEMP
		movwf	INDF		; Value now in array
		incf	SCRATCH+1,F
		incf	SCRATCH+1,F	; srcptr = srcptr + 2
		incf	SCRATCH+2,F	; dstptr = dstptr + 1
		decfsz	SCRATCH,F	; loop 32 samples
		 goto	_subsample2

		clrc
		rlf	Interval,F	; Double the interval
		rrf	SampleCnt,F	; Halve the sample count
		
_doStore	btfsc	SampleCnt,6
		 goto	storeHighBank
storeLowBank	movfw	SampleCnt
		addlw	SampleArrLow
		movwf	FSR
		movfw	ActVoltage	; Get Last measured value (0..255)
		movwf	INDF		; Store new sample
		incf	SampleCnt,f	; Increment nr of Sample
		return
storeHighBank	movfw	SampleCnt
		andlw	0x3F		; Mask of high bit
		addlw	low SampleArrHigh
		movwf	FSR
		movfw	ActVoltage	; Get Last measured value (0..255)
		bsf	STATUS,IRP	; Select high banks
		movwf	INDF		; Store new sample
		bcf	STATUS,IRP
		incf	SampleCnt,f	; Increment nr of Sample
		return

;*****************************************************************************
; Send the contents of the SampleArray over a (bit-banged) serial interface
;*****************************************************************************
SendSamples:	bsf	E_SendSerial	; First send capacity (in mAh)
		call	DispCapacity
		bcf	E_SendSerial
		call	SendCRLF

		clrf	TEMP		; Next send discharge curve
SendLoop:	movfw	TEMP
		movwf	(SCRATCH)
		clrf	(SCRATCH+1)
		call	BIN2ASC		; convert counter to ASCII
		movf	hund,w
		call	SEROUT
		movf	ten,w
		call	SEROUT
		movf	one,w
		call	SEROUT
		SNDCHAR	','

		btfsc	TEMP,6
		 goto	ReadHighBank
ReadLowBank	movfw	TEMP
		addlw	SampleArrLow
		movwf	FSR
		movfw	INDF		; Read sample from memory
		goto	_send
ReadHighBank	movfw	TEMP
		andlw	0x3F		; Mask of high bit
		addlw	low SampleArrHigh
		movwf	FSR
		bsf	STATUS,IRP	; Select high banks
		movfw	INDF		; Read sample from memory
		bcf	STATUS,IRP
_send		bsf	E_SendSerial
		call	DispVoltage
		bcf	E_SendSerial
		call	SendCRLF
		incf	TEMP,f
		movfw	TEMP
		subwf	SampleCnt,w
		bnz	SendLoop
		return

;DispScreen	call	DispScreen1	; Dispscreen is only used on 16*2 displays
;		movlw	LCD_LINE1 + 2
;		call	LCDSDDA
;		call	DispScreen2
;		return

DispScreen1:	movf	ActVoltage,w
		call	DispVoltage

		call	CalcCurrent
		call	DispCurrent
		movlw	' '
		call	LCDPUTCHAR
		return

DispScreen2:	movf	HOURS,w
		call	DispByte2
		movlw	':'
		call	LCDPUTCHAR
		movf	MINS,w
		call	DispByte2
		movlw	' '
		call	LCDPUTCHAR
		call	DispCapacity
		return		

DispByte2:	movwf	(SCRATCH)
		clrf	(SCRATCH+1)
		call	BIN2ASC		; convert counter to ASCII
		movf	ten,w
		call	LCDPUTCHAR	; Display character
		movf	one,w
		call	LCDPUTCHAR	; Display character
		return

ShowMsg:	movwf	TEMP
		call	LCDCLEAR
		movfw	TEMP
		BANK2
		movwf	EEADR		; Data Memory Address
MsgLoop:	BANK3	
		bcf	EECON1,EEPGD	; Point to Data memory
		bsf	EECON1,RD	; EE Read
		BANK2
		incf	EEADR,F		; Point to next location
		movf	EEDATA,w 	; W = EEDATA
		BANK0
		skpnz			; At end of string ?
		 return
		call	LCDPUTCHAR
		goto	MsgLoop

GetResistor:	movlw	LCD_LINE0+.7	; Position Cursor
		call	LCDSDDA
		clrc
		rrf	Resistor,W	; Get resistor /2 (whole part)
		call	DispByte3
		movlw	'.'
		call	LCDPUTCHAR
		movlw	'0'
		btfsc	Resistor,0	; Test for fraction: 0 or 1/5
		movlw	'5'
		call	LCDPUTCHAR
		bcf	E_Tick		; Clear Event
_scnloop:	btfsc	E_But1Pressed
		 return
		btfss	But2State	; Look at filtered state
		 goto	_scnloop
		btfss	E_Tick
		 goto	_scnloop
		 
		bcf	E_Tick
		movlw	.250		; Resistor Range is 1..125 Ohm
		subwf	Resistor,w
		skpnz
		 clrf	Resistor
		incf	Resistor,f	; While button is down increase val every second
		goto	GetResistor

DispByte3:	movwf	(SCRATCH)
		clrf	(SCRATCH+1)
		call	BIN2ASC		; convert counter to ASCII
		movf	hund,w
		call	LCDPUTCHAR	; Display character
		movf	ten,w
		call	LCDPUTCHAR	; Display character
		movf	one,w
		call	LCDPUTCHAR	; Display character
		return

GetVRange:	call	DispVRange
		bcf	E_Tick
_scnloop3:	btfsc	E_But1Pressed
		 return
		btfss	But2State	; Look at filtered state
		 goto	_scnloop3
		btfss	E_Tick
		 goto	_scnloop3
		 
		incf	V_Range,W	; While button is down increase val every second
		andlw	0x03		; 4 Possible ranges for VRange
		movwf	V_Range
	
		bcf	E_Tick
		goto	GetVRange
		
DispVRange:	movlw	LCD_LINE0+.10	; Position Cursor
		call	LCDSDDA
		call	CalcVScale	; Range could be changed, so update VSCALE
		movlw	.255		; Display maxval
		call	DispVoltage
		return

;*****************************************************************************
; Calculate scaling factor from the A/D input to VRange (in milliVolts)
; (The scaling factor depends on the resistor divider (Max Range) used
;*****************************************************************************
CalcVScale:	FPLDC24	.255		; Voltage in Milivolts: (X / 255) * VRange 
		call	FPMOV24		; So scaling fac is (Vref / 255)

VR_Table	movlw	high $		; Load High part of program counter
		movwf	PCLATH
         	movfw	V_Range
		andlw   0x03            ; limit to [0..3]
		addwf   PCL,F           ; offset in jump Table
		goto	Range25		;  2.5 V Range
		goto    Range50		;  5.0 V Range
		goto	Range75		;  7.5 V Range
		goto	Range150	; 15.0 V Range
VR_TableEnd
    IF HIGH(VR_Table) != HIGH(VR_TableEnd) 
        error "Voltage Range Table crosses page border"
    ENDIF

Range25		FPLDC24	.2500		;  2.5 V reference
		goto	_calcvr
Range50		FPLDC24	.5000		;  5.0 V reference
		goto	_calcvr
Range75		FPLDC24	.7500		;  7.5 V reference
		goto	_calcvr
;Range150	FPLDC24	.15000		; 15.0 V reference
Range150	FPLDC24	.12500		; 12.5 V reference
		goto	_calcvr

_calcvr		clrf	PCLATH		; Other routines expect this to be 0
		call	FPDIV24		; Multiply by Vref to give range 0..VRef*1000 (milliVolts)
		FPSTF24	VScale
		return

GetCOVoltage:	call	DispCOVoltage
		bcf	E_Tick
_scnloop2:	btfsc	E_But1Pressed
		 return
		btfss	But2State	; Look at filtered state
		 goto	_scnloop2
		btfss	E_Tick
		 goto	_scnloop2
		 
		incf	CO_Voltage,f	; While button is down increase val every second
		bcf	E_Tick
		goto	GetCOVoltage
		
DispCOVoltage:	movlw	LCD_LINE0+.10	; Position Cursor
		call	LCDSDDA
		
		movf	CO_Voltage,w
		call	DispVoltage
		return

;*****************************************************************************
; Display a Voltage, The sample range runs from 0..255 for 2.50 V (ref. diode)
; With a suitable divider the range can be set to other values like 0..15 V
;*****************************************************************************
DispVoltage:	movwf	AARG+1
		clrf	AARG
		call	FPFLO24		; W   -> FPA
		call	FPMOV24		; FPA -> FPB
		FPLDF24	VScale
		call	FPMUL24		; Multiply by eg 5000/255 to get a 0..5000mV range
		call	FPINT24		; convert to integer

		movf	AARG+1,w
		movwf	SCRATCH
		movf	AARG,w
		movwf	SCRATCH+1
		
		call	BIN2ASC

		btfsc	E_SendSerial	; SendSerial defines whether the data should be
		 goto	SendVoltage	; shown on the LCD or send via the RS232 iface

		movf	tenk,w
		call	LCDPUTCHAR
		movf	onek,w
		call	LCDPUTCHAR
		movlw	'.'
		call	LCDPUTCHAR
		movf	hund,w
		call	LCDPUTCHAR
		movf	ten,w
		call	LCDPUTCHAR
		movlw	'V'
		call	LCDPUTCHAR
		return

SendVoltage:	movf	tenk,w		; Send 16 value from BIN2BCD routine
		call	SEROUT		; over the serial interface
		movf	onek,w		; format as "XX.XX" (like 10.00)
		call	SEROUT
		SNDCHAR	'.'
		movf	hund,w
		call	SEROUT
		movf	ten,w
		call	SEROUT
		return
		
CalcCurrent:	movf	ActVoltage,W	; Calculate the current using the value
		movwf	AARG+1		; of the load resistor and the voltage
		clrf	AARG
		call	FPFLO24
		call	FPMOV24
		FPLDF24	VScale
		call	FPMUL24
		call	FPMOV24		; Voltage in milliVolts in FPB

		movf	Resistor,w	; Note: this are ohms * 2
		movwf	AARG+1
		clrf	AARG
		call	FPFLO24		; 2 * Resistance in FPA
		call	FPSWP24
		call	FPDIV24		; .5 * Current in FPA
		call	FPSWP24		; Now in FPB
		FPLDC24	.2
		call	FPMUL24		; Multiply by 2 (
		FPSTF24	Current
		return
		
DispCurrent:	FPLDF24	Current
		call	FPINT24

		movf	AARG+1,w
		movwf	SCRATCH
		movf	AARG,w
		movwf	SCRATCH+1
		
		call	BIN2ASC
		movlw	' '
		call	LCDPUTCHAR
		movf	onek,w
		call	LCDPUTCHAR
		movf	hund,w
		call	LCDPUTCHAR
		movf	ten,w
		call	LCDPUTCHAR
		movf	one,w
		call	LCDPUTCHAR
		
		movlw	'm'
		call	LCDPUTCHAR
		movlw	'A'
		call	LCDPUTCHAR
		return
		
DispCapacity:	FPLDC24	.60			; Divide the capacity by 60 to get mAH
		call	FPMOV24
	  	FPLDF24	Capacity
	  	call	FPDIV24
		call	FPINT24			; mAH now available as INT

		movf	AARG+1,w
		movwf	SCRATCH
		movf	AARG,w
		movwf	SCRATCH+1
		
		call	BIN2ASC		

		btfsc	E_SendSerial	; SendSerial defines whether the data should be
		 goto	SendCapacity	; shown on the LCD or send via the RS232 iface

		movf	tenk,w
		call	LCDPUTCHAR
		movf	onek,w
		call	LCDPUTCHAR
		movf	hund,w
		call	LCDPUTCHAR
		movf	ten,w
		call	LCDPUTCHAR
		movf	one,w
		call	LCDPUTCHAR
		movlw	'm'
		call	LCDPUTCHAR
		movlw	'A'
		call	LCDPUTCHAR
		movlw	'h'
		call	LCDPUTCHAR
		return

SendCapacity	movfw	tenk		; Send 16 value from BIN2BCD routine
		call	SEROUT		; over the serial interface
		movfw	onek		; format as "XXXXX" (like 10000)
		call	SEROUT
		movfw	hund
		call	SEROUT
		movfw	ten
		call	SEROUT
		movfw	one
		call	SEROUT
		return

;*****************************************************************************
; Increment the battery capacity
; The capacity is held in a 24 bit float so do not update to often as the
; dynamic range of the float is only 4.6 digits
; (So of the capacity increases the resolution will degrade on many updates)
; Please note that internally the capacity is kept in milliAmpereMinutes
;*****************************************************************************
AdjustCapacity:	FPLDF24	Current
		call	FPMOV24
		FPLDF24	Capacity
		call	FPADD24
		FPSTF24	Capacity
       	        return
                
; Measure the Voltage on AN0

Measure:	bsf	ADCON0,GO	; Start conversion
		btfsc	ADCON0,NOT_DONE	; Wait for convrersion to complete (+/- 20usec)
		 goto	$-1
		 
		movf	ADRESH,w	; Get result
		movwf	ActVoltage
		return

LoadBarChars:	clrw			; Point to character ram
		call	LCDSCGA
		movlw	8
		movwf	TEMP2		; Characters to load

_loop0		decf	TEMP2,w		; First load the "white" lines
		movwf	TEMP
		bz	_doblack
_loop1		movlw	0
		call	LCDPUTCHAR
		decfsz	TEMP,f
		goto	_loop1

_doblack	movf	TEMP2,w		; then load the "black" lines
		sublw	8
		movwf	TEMP
		bz	_dounderscore
_loop2		movlw	0x1F
		call	LCDPUTCHAR
		decfsz	TEMP,F
		 goto	_loop2
		
_dounderscore	movlw	0
		call	LCDPUTCHAR
		decfsz	TEMP2,f
		 goto	_loop0
		return

CheckPower:	bsf	ADCON0,CHS0	; Select AN1
		movlw	.255		; Acquisition delay
		movwf	SCRATCH
		decfsz	SCRATCH,F
		 goto	$-1
		bsf	ADCON0,GO	; Start conversion
		btfsc	ADCON0,NOT_DONE	; Wait for conversion to complete (+/- 20usec)
		 goto	$-1

		bcf	ADCON0,CHS0	; Select AN0 again (Sense input)
		movf	ADRESH,W	; Get battery result
		return

;*****************************************************************************
; The interrupt handling is used for time keeping (Maintains system clock)
; and also used for switch debouncing. An event flag will be raised after a
; down/up sequence
;*****************************************************************************

INTHANDLER	movwf	_w		; Save W and STATUS
		movf	STATUS,W
		movwf	_status
		bcf     STATUS, RP0	; Select bank 0
		bcf	STATUS, RP1	

		bcf 	PIR1, TMR2IF	; Clr Timer2 interrupt

					; We count to 50 using to counters:
					; Inner counts to 10, outer to 5
					; This will give an 200 msec Event
					; flag which is used for repeat
		decfsz	TICKS1,f
		 goto	doKeys
		bsf	E_Tick		; raise 200msec event
		movlw	.10		; reset tick counter
		movwf	TICKS1
		decfsz	TICKS2,f
		 goto	doKeys	
		movlw	.5
		movwf	TICKS2
		bsf	E_Second	; raise 1 second event
		incf	SECS,f
		movlw	.60
		subwf	SECS,w
		skpz 
		 goto 	doKeys
		bsf	E_Minute	; raise 1 minute event
		clrf	SECS
		incf	MINS,F
		movlw	.60
		subwf	MINS,W
		skpz
		 goto 	doKeys
		clrf	MINS
		incf	HOURS,F		; increment hours

		; Get Switch status and place in csa  bit 0 and 1
		; (Allow remapping)
		
doKeys		clrf	DBNC_csa
		btfss	But1Sense
		 bsf	DBNC_csa,0
		btfss	But2Sense
		 bsf	DBNC_csa,1

		call	DeBounce

		btfss	cva_old,0
		 goto	_L0
		btfss	DBNC_cva,0
		 bsf	E_But1Pressed
		 
_L0 		btfss	cva_old,1
		 goto	_L1
		btfss	DBNC_cva,1
		 bsf	E_But2Pressed

_L1		movf	DBNC_cva,w
		movwf	cva_old

int_done	movf	_status,w	; Restore W and STATUS
		movwf	STATUS
		swapf	_w,F
		swapf	_w,W
		retfie

SendEEData	bcf	E_But2Pressed
		call	LoadDischarge
		movf	SampleCnt,f
		bz	_noEEdata	; No Discharge Data available
		call	CalcVScale	; Calculcate V-scaling to use
		call	SendSamples	; And send the data
		return

_noEEdata	SNDCHAR	'N'
		SNDCHAR	'o'
		SNDCHAR	' '
		SNDCHAR	'D'
		SNDCHAR	'a'
		SNDCHAR	't'
		SNDCHAR	'a'

SendCRLF	SNDCHAR	.13
		SNDCHAR	.10
		return
		
; 16 step 8 bit averaging filter
; The formula can be rewritten as: Ave' = Ave + ( New - Ave )/16 ; 
;
; Used the same names for variables. TEMP cell isn't used. 
; 
; I having more and more believing that 12 is some magical number for small completed code fragments. ;) 
;  Initially written by Andrew Warren (fastfwd at ix.netcom.com).
;  Optimized by Dmitry A. Kiryashov (zews at aha.ru) 06/18/2000
;  12 clocks/words
;
;Filter:
;        movfw   AVE
;        subwf   NEW,F   ;NEW - AVE
;
;        swapf   NEW,W
;        andlw   0x0F    ;get /16 int part
;        skpc            ;result is neg?
;        iorlw   0xF0    ;yes
;        addwf   AVE,F
;
;        swapf   NEW,W
;        andlw   0xF0    ;get /16 frac part
;        addwf   AVFRAC,F
;        skpnc
;        incf    AVE,F
;
;       low nibble of AVFRAC isn't used
;       so we can utilize it for other purposes ;)

;*****************************************************************************
; Keep user settings in EE memory
; Location 0:	Resistor Load		(1..240)
; Location 1: 	Cutoff Voltage		(0..255) (should be scaled !!)
; Location 2:	Sampling Interval Index	(0.3)    NOT USED ANYMORE
; Location 3:	Voltage Range		(0.3)    (Index in table) 
;*****************************************************************************
LoadPars:	movlw	Resistor
		movwf	FSR		; FSR to first variable
		BANK2
		clrf	EEADR		; Data Memory Address
		BANK0
		movlw	.4
		call	LoadEE
		return		

SavePars:	movlw	Resistor	; Store 4 bytes starting from Resistor
		movwf	FSR
		BANK2
		clrf	EEADR		; Starting at loc. 0 in EE
		BANK0
		movlw	.4
		call	SaveEE
		return

;*****************************************************************************
; LoadDischarge restores the last complete discharge cycle from EE
; The Capacity is kept at location 0x217D [  3bytes]
; The Curve    is kept at location 0x2180 [128bytes]
;*****************************************************************************
LoadDischarge:	movlw	Capacity	; Save capacity & sample count
		movwf	FSR
		BANK2
		movlw	low EECapacity
		movwf	EEADR		; Starting location in EE
		BANK0
		movlw	.4		; 3 bytes cap + 1 byte cnt
		call	LoadEE

		movlw	SampleArrLow	; Load curve
		movwf	FSR		; first 64 locations
		BANK2
		movlw	low EESamples
		movwf	EEADR		; Starting location in EE
		BANK0
		movlw	.64		; Samples to store
		call	LoadEE

		movlw	low SampleArrHigh
		movwf	FSR		; second 64 locations
		bsf	STATUS,IRP	; Select high banks for FSR
		BANK2
		movlw	low EESamples + 0x40
		movwf	EEADR		; Starting location in EE
		BANK0
		movlw	.64		; Samples to store
		call	LoadEE
		bcf	STATUS,IRP	; Select low banks for FSR
		return

;*****************************************************************************
; SaveDischarge saves the last complete discharge cycle in EE
;*****************************************************************************
SaveDischarge:	movlw	Capacity	; Save capacity & sample count
		movwf	FSR
		BANK2
		movlw	low EECapacity
		movwf	EEADR		; Starting location in EE
		BANK0
		movlw	.4		; 3 bytes cap + 1 byte cnt
		call	SaveEE

		movlw	SampleArrLow	; Save curve
		movwf	FSR		; first 64 locations
		BANK2
		movlw	low EESamples
		movwf	EEADR		; Starting location in EE
		BANK0
		movlw	.64		; Samples to store
		call	SaveEE

		movlw	low SampleArrHigh
		movwf	FSR		; second 64 locations
		bsf	STATUS,IRP	; Select high banks for FSR
		BANK2
		movlw	low EESamples + 0x40
		movwf	EEADR		; Starting location in EE
		BANK0
		movlw	.64		; Samples to store
		call	SaveEE
		bcf	STATUS,IRP	; Select low banks for FSR
		return

;*****************************************************************************
; LoadEE loads WREG bytes starting from EE memory pointed to by EEADR
; into the location pointed to by FSR
;*****************************************************************************
LoadEE:		movwf	TEMP
_loadloop	BANK3	
		bcf	EECON1,EEPGD	; Point to Data memory
		bsf	EECON1,RD	; EE Read
		BANK2
		movf	EEDATA,w 	; W = EEDATA
		movwf	INDF
		incf	FSR,F
		incf	EEADR,f		; To next EE Address
		BANK0
		decfsz	TEMP,f
		 goto	_loadloop
		return		

;*****************************************************************************
; SaveEE saves WREG bytes starting from the location pointed to by FSR
; into EE memory pointed to by EEADR
;*****************************************************************************
SaveEE:		movwf	TEMP		; Variable count
		BANK3
		bcf	EECON1,EEPGD	; Point to Data memory
		bsf	EECON1,WREN	; Enable EE write
_saveloop:	BANK2
		movf	INDF,w
		movwf	EEDATA
		BANK3
		call	EEWrite
		BANK2
		incf	EEADR,f
		incf	FSR,F
		BANK0
		decfsz	TEMP,f
		 goto	_saveloop
		BANK3
		bcf	EECON1,WREN	; Disable EE write
		BANK0
		return
		
	ERRORLEVEL	-302		
EEWrite:	bcf	INTCON,GIE	; Disable interrupts
		movlw	0x55		; Write required sequence
		movwf	EECON2
		movlw	0xAA
		movwf	EECON2
		bsf	EECON1,WR
		bsf	INTCON,GIE	; Enable interrupts again
		btfsc	EECON1,WR	; wait for write to complete ( typ. 10 mSec)
		 goto	$-1
		return
	ERRORLEVEL	+302		

;*****************************************************************************
; EEPROM contents: Keeps user variables and display messages
;		   Keeps "historical" data (Curve & Capacity)
;*****************************************************************************

		org 0x2100		; Room for 8 User variables
		de .100			; Load Resistor:	50 ohm (100/2)
		de .153			; Cutoff Voltage:	3.00 V
		de .5			; Sampling interval:	NOT USED ANYMORE
		de .1			; Voltage range:	5.0 V (entry 2 in table)
		
		org 0x2108		; Start of Text area
msgVersion	de "Cap. Meter V 1.6",0	; User Messages
msgSetResistor	de "Load R xxx.x ohm",0
msgSetCutOff	de "V Cutoff",0
msgSetVRange	de "V Range",0
msgShowGo	de "xx.xxV  Press GO",0
msgLowBat	de "* Battery low *",0
msgShowRdy	de "Ready   ",0

		org 0x217C
EECapacity	de 0,0,0		; EE copy of Capacity
EESampleCnt	de 0			; EE copy of nr of samples
EESamples				; EE copy of Discharge Curve (128 bytes)

	END				; End of program
