;*****************************************************************************
;
; Subroutine to convert a 2 byte unsigned integer into a 5 byte decimal value
; please note the input value is destroyed in the conversion process
;
; Program	BIN2ASC.INC
; Author	Wim de Vree, PE1GRL@PI8WNO	
; Revision history
;	990611	Created, based upon David Bensons example (Pic'n up the Pace pg 233-238)
;	990612	Rewritten as a INC file (conversion subroutine only)
;
; Usage
;	
;
; Example
;		SCRATCH 	equ 0x10	; temp workspace, 7 bytes required
;		B2A_RESULT	equ 0x20	; result will be places here,
;						; 5 bytes required, high to low order
;
;		movlw	0xFA			; Input LSB
;		movwf	(SCRATCH)
;		movlw	0x00			; Input MSB
;		movwf	(SCRATCH+1)
;		call	BIN2ASC

;*****************************************************************************
; Equates used by subroutines
;*****************************************************************************
;SCRATCH		equ	0x0C
lsb2		equ	(SCRATCH)	; high byte binary input  value
msb2		equ	(SCRATCH+1)	; low  byte binary output value
index		equ	(SCRATCH+2)
dig_ctr		equ	(SCRATCH+3)
flags		equ	(SCRATCH+4)
lsb1		equ	(SCRATCH+5)
msb1		equ	(SCRATCH+6)
;B2A_RESULT	equ	0x28
tenk		equ	(B2A_RESULT)
onek		equ	(B2A_RESULT+1)
hund		equ	(B2A_RESULT+2)
ten		equ	(B2A_RESULT+3)
one		equ	(B2A_RESULT+4)
; bit equates
ovflw		equ	0

tbl_lo		addwf	PCL,f		; add index to program counter
		retlw	0x10		; 10,000 decimal
		retlw	0xe8		;  1,000 decimal
		retlw	0x64		;    100 decimal
		retlw	0x0a		;     10 decimal
;-----------------------------------------------------------------------
tbl_hi		addwf	PCL,f		; add index to program counter
		retlw	0x27		; 10,000 decimal
		retlw	0x03		;  1,000 decimal
		retlw	0x00		;    100 decimal
		retlw	0x00		;     10 decimal
;-----------------------------------------------------------------------
BIN2ASC
		movlw	0x30		; init result with ASCII '0'
		movwf	tenk
		movwf	onek
		movwf	hund
		movwf	ten
		movwf	one
		clrf	index		; init work registers
		clrf	dig_ctr
		bcf	flags,ovflw
subtr		movf	index,w		; get current index into W
		call	tbl_lo		; get ls chunk for subtraction
		movwf	lsb1
		movf	index,w		; get current index into W
		call	tbl_hi		; get ms chunk for subtraction
		movwf	msb1
		call	dblsub		; double precision subtraction
		btfsc	STATUS,C	; test carry flag
		goto	incdig
		movlw	B2A_RESULT	; load base address of table
		movwf	FSR
		movf	index,w		; get index
		addwf	FSR,f		; add offset
		movf	dig_ctr,w	; get digit counter contents
		addwf	INDF,f		; add to '0' at digit loc (indexed)
		movf	index,w		; get index
		call	tbl_lo		; get ls chunk for addition
		movwf	lsb1
		movf	index,w		; get index
		call	tbl_hi		; get ms chunk for addition
		movwf	msb1
		bcf	flags,ovflw
		call	dblplus		; to double precision addition
		movf	index,w		; get index
		sublw	0x03		; index = 3 ?
		btfsc	STATUS,Z
		goto	finish
		incf	index,f
		clrf	dig_ctr
		goto	subtr
;------------------------------------------
incdig		incf	dig_ctr,f	; increment digit counter
		goto	subtr
;------------------------------------------
finish		movf	lsb2,w		; get 1's = remainder
		addwf	one,f		; add to '0'
		return			; done
;------------------------------------------
dblsub		bcf	flags,ovflw	; clear overflow flag
		comf	lsb1,f		; 2's complement stuff
		comf	msb1,f
		movf	lsb1,w
		addlw	0x01
		movwf	lsb1
		btfsc	STATUS,C
		incf	msb1,f
dblplus		movf	lsb1,w
		addwf	lsb2,f
		btfsc	STATUS,C
		bsf	flags,ovflw	; indicate overflow occurred
		movf	msb1,w
		addwf	msb2,f		; add high bytes
		btfss	flags,ovflw
		goto 	dblx
		btfsc	STATUS,C
		goto	db11
		movlw	0x01
		addwf	msb2,f
		goto	dblx
db11		incf	msb2,f
dblx		return

