;*****************************************************************************
;
; Subroutine to divide 16 bits by 8 bits, 8 bit result
; uses conventional shift and subtract
; The result MUST fit in 8 bits, otherwise erroneous result. 
; requires 110 clock cycles. Originally Written by: Bob Fehrenbach
;
; Program	DIV168.INC
; Author	Wim de Vree, PE1GRL@PI8WNO	
; Revision history
;	991121	Downloaded from the Internet 
;
; Used resources:
;		SCRATCH		< 5 bytes>
;
; Usage:
;		SCRATCH 	equ 0x10	; workspace, 3 bytes required
;						; place input at scratch
;						; result will be placed here as well
;
;		movlw	0xFA			; Input MSB
;		movwf	(SCRATCH)
;		call	BYT2ASC
;
;
;
;*****************************************************************************
; Equates used by subroutines
;*****************************************************************************

__dvdLO		equ	(SCRATCH)	; high byte binary input  value
__dvdHI		equ	(SCRATCH+1)
__dvsrLO	equ	(SCRATCH+2)	; divisor
__dvdCOUNT	equ	(SCRATCH+3)	; temp variable
__tmpLO		equ	(SCRATCH+4)	; temp variable

DIV168:  	movlw   8 
		movwf   __dvdCOUNT
div_loop: 
		rlf     __dvdLO, F 
		rlf     __dvdHI, F 
		rlf     __tmpLO, F      ;save "borrow" 
		movf    __dvsrLO, W 
		subwf   __dvdHI, W 
		btfss   __tmpLO, 0      ;Save result if borrow = 1 
		skpnc                   ;Else, save only if positive 
		movwf	__dvdHI 
		btfsc   __tmpLO, 0      ;Make sure carry is a 1 if 
		rrf     __tmpLO, F	;borrow = 1 
		decfsz  __dvdCOUNT, F 
		goto    div_loop 
		rlf     __dvdLO, F	;quotient in dvdLO 
div_exit:				;remainder in dvdHI 
		return 
;*****************************************************************************
