;*****************************************************************************
;
; SERINIT:	Set up the serial output pin
;
; SEROUT:	Subroutine to transmit serial data (1stop/8bit) 
;		Data in 'W' is transmitted.
;
; Program	SEROUT.INC
; Author	Wim de Vree, pe1grl@amsat.org
; Revision history
; 1999-11-06	Created,
;		based upon code from Peter H. Anderson, MSU, May 29, '97
; 1999-12-26	More "portable", you can define PORT,BIT and BAUDRATE
; 1999-12-28	Added init routine to set the port properties
; 2000-02-20	Bug fixed: stop bit had constant length (so too long for hi-bauds)
;		Added 4800 & 9600 Baud constants
; 2000-05-07	Added 4.096 MHz X-tal support, use MACROS.INC
; 2002-12-08	Bug fixed in defines for 4.096 MHz
; 2003-04-09	Added 1000 Baud (used on RF link)
; 2004-10-24	Define CLOCK_FREQ with decimal values
; 2004-11-08	BaudRates are defined in decimal
;		Added SNDCHAR macro
;
; Used resources:
;		SCRATCH		< 3 bytes>
;
; Usage:
;		SCRATCH 	equ 0x10	; workspace, 3 bytes required
;
;#define	
;#define	BD_DELAY	1200BD_DELAY	; selected Baud Rate
;#define MAXIM_USED	1		; MAXIM232 RS232 used (inverting!)
;#define SERPORT	PORTA		; port used for serial I/O
;	 CONSTANT 	TXBIT=4		; bit used for TX
;
;		movlw	'A'			; Character to transmit
;		call	SEROUT
;
;
;
;*****************************************************************************
; Equates used by subroutines
;*****************************************************************************

	if	  CLOCKFREQ == .18432
	CONSTANT	DELAY1200BD=.125
	CONSTANT	DELAY2400BD=.61
	CONSTANT	DELAY4800BD=.29
	CONSTANT	DELAY9600BD=.13
	else
	    if	  CLOCKFREQ == .40000
	CONSTANT	DELAY0600BD=.207	; 5 NOPS in loop !!!!
	CONSTANT	DELAY1000BD=.248	; 1 NOP in loop
	CONSTANT	DELAY1200BD=.206	; 1 NOP in loop
	CONSTANT	DELAY2400BD=.102	; 1 NOP in loop
	CONSTANT	DELAY4800BD=.50		; 1 NOP in loop
;	CONSTANT	DELAY2400BD=.135	; NO NOP NEEDED
;	CONSTANT	DELAY4800BD=.66		; NO NOP NEEDED
;	CONSTANT	DELAY9600BD=.32		; NO NOP NEEDED
	    else
	       if CLOCKFREQ == .40960
	CONSTANT	DELAY1200BD=.211	; NOP NEEDED !!!!
	CONSTANT	DELAY2400BD=.104	; NOP NEEDED
	CONSTANT	DELAY4800BD=.51		; NOP NEEDED
	       else
	ERROR	"SEROUT: No ClockFrequency defined"
	       endif
	   endif
	endif

 	if			BDRATE ==  .600
#define BD_DELAY		DELAY0600BD
	else 	
 	    if			BDRATE == .1000
#define BD_DELAY		DELAY1000BD
	    else 	
 	    	if		BDRATE == .1200
#define BD_DELAY		DELAY1200BD
	    	else
		    if		BDRATE == .2400
#define BD_DELAY		DELAY2400BD
		    else
		    	if 	BDRATE == .4800
#define BD_DELAY		DELAY4800BD
		        else
		            ERROR	"SEROUT: No or Invalid Baud Rate defined"
		    	endif
	            endif
	    	endif
	    endif
	endif
 
TXONE	MACRO
	ifdef   MAXIM_USED
		bsf	SERPORT, TXBIT	; Note:MAXIM is inverting chip
	else
		bcf	SERPORT, TXBIT
	endif
	ENDM

TXZERO	MACRO
	ifdef   MAXIM_USED
		bcf	SERPORT, TXBIT
	else
		bsf	SERPORT, TXBIT
	endif
	ENDM
	
SNDCHAR	MACRO	CHAR
	movlw	CHAR
	call	SEROUT
	ENDM

SEROUT_DATA	equ	(SCRATCH)	; Sign placed here (00 = pos, FF = neg)
SEROUT_LOOP	equ	(SCRATCH+1)	;
SEROUT_TIME	equ	(SCRATCH+2)	;
;
; set the serial pin to the proper idle setting
;
SERINIT:
		TXONE			; Pull to logic "1" level
		return
;
; transmits content of W at the desired baudrate
;
SEROUT:
		movwf	SEROUT_DATA
		movlw 	.9
		movwf	SEROUT_LOOP
		bcf	STATUS, C	; set C to 0, start bit

SEROUT1:
		btfsc	STATUS, C	; 1 cycles		; 1
		TXONE			; transmit a `one'	; 2a
		btfss	STATUS, C				; 3
		TXZERO			; or a 'zero' 		; 2b
		movlw   BD_DELAY	; bit time delay,	; 4
	        movwf	SEROUT_TIME     ; 814uS for 1200 baud	; 5
SEROUT2:
	if	CLOCKFREQ >= .40000
		nop
	endif
	if	BDRATE	< 1000
		nop
		nop
		nop
		nop
	endif
		decfsz  SEROUT_TIME, F	; 1 cycles in loop
	        goto	SEROUT2		; 2 CYCLES IN LOOP
	
		rrf	SEROUT_DATA, F	; leastsignbit now in C	; 6
		decfsz	SEROUT_LOOP, F	; does not affect status; 7
		goto 	SEROUT1		; next character	; 8 + 9
		
		TXONE			; send stop bit
		movlw   BD_DELAY
	        movwf	SEROUT_TIME    	; bit time delay, 814uS for 1200 baud
SEROUT3:
	if	CLOCKFREQ >= .40000
		nop
	endif
	if	BDRATE	< 1000
		nop
		nop
		nop
		nop
	endif
		decfsz  SEROUT_TIME, F
	        goto	SEROUT3

		return

;------------------------------------------------------------------------
;END SEROUT ROUTINES
;------------------------------------------------------------------------
